import json

sample_buffer_size = 16384


def encode_string(string):
    return (string + "\0").encode("utf-8")


def allocate_array_on_stack(instance, arr):
    pointer = instance.exports.stackAlloc(len(arr))

    HEAP8 = instance.exports.memory.uint8_view(offset=pointer)

    for i, c in enumerate(arr):
        HEAP8[i] = c

    return pointer, lambda: HEAP8[0: len(arr)]


def allocate_profile_on_stack(instance, profile):
    cProfiles, _ = allocate_array_on_stack(
        instance,
        encode_string(
            json.dumps({
                "profile": profile
            })
        )
    )

    cProfile, _ = allocate_array_on_stack(
        instance,
        encode_string('profile')
    )

    return cProfile, cProfiles



from .utils import *
import pyaudio


def chunks(arr, n):
    for i in range(0, len(arr), n):
        yield arr[i:i + n]


FORMAT = pyaudio.paFloat32
CHANNELS = 1
SAMPLE_RATE = 48000

p = pyaudio.PyAudio()
stream = p.open(format=FORMAT,
                channels=CHANNELS,
                rate=SAMPLE_RATE,
                output=True)


class Transmitter:
    def __init__(self, instance):
        self.destroyed = False
        self.instance = instance

    def select_profile(self, profile):
        stack = self.instance.exports.stackSave()
        cProfile, cProfiles = allocate_profile_on_stack(self.instance, profile)

        quiet_encoder_options = self.instance.exports.quiet_encoder_profile_str(
            cProfiles,
            cProfile
        )

        self.encoder = self.instance.exports.quiet_encoder_create(
            quiet_encoder_options,
            float(SAMPLE_RATE)
        )

        self.frame_length = self.instance.exports.quiet_encoder_clamp_frame_len(
            self.encoder,
            sample_buffer_size
        )

        self.instance.exports.free(quiet_encoder_options)
        self.instance.exports.stackRestore(stack)
        return self

    def transmit(self, buf):
        stack = self.instance.exports.stackSave()

        audioSampleBytesPointer, getAudioSampleBytes = allocate_array_on_stack(
            self.instance,
            [0] * sample_buffer_size * 4
        )

        payload = chunks(buf, self.frame_length)

        pre_loop_stack = self.instance.exports.stackSave()
        for frame in payload:

            unicodeBytesPointer, _ = allocate_array_on_stack(
                self.instance,
                frame
            )

            self.instance.exports.quiet_encoder_send(
                self.encoder,
                unicodeBytesPointer,
                len(frame)
            )

            self.instance.exports.quiet_encoder_emit(
                self.encoder,
                audioSampleBytesPointer,
                sample_buffer_size
            )

            play_bytes(bytes(getAudioSampleBytes()))
            self.instance.exports.stackRestore(pre_loop_stack)

        self.instance.exports.stackRestore(stack)
        return self

    def destroy(self):
        if (not self.destroyed):
            self.instance.exports.quiet_encoder_destroy(self.encoder)
            self.destroyed = True
        return self


def play_bytes(raw_bytes):
    data = raw_bytes
    stream.write(data)



import pyaudio
from queue import Queue
from .utils import *

FORMAT = pyaudio.paFloat32
CHANNELS = 1
SAMPLE_RATE = 48000

p = pyaudio.PyAudio()


class Receiver:
    def __init__(self, instance):
        self.destroyed = False
        self.instance = instance

    def select_profile(self, profile):
        stack = self.instance.exports.stackSave()
        cProfile, cProfiles = allocate_profile_on_stack(self.instance, profile)

        quiet_decoder_options = self.instance.exports.quiet_decoder_profile_str(
            cProfiles,
            cProfile
        )

        self.quiet_decoder = self.instance.exports.quiet_decoder_create(
            quiet_decoder_options,
            float(SAMPLE_RATE)
        )

        self.instance.exports.free(quiet_decoder_options)
        self.instance.exports.stackRestore(stack)
        return self

    def receive(self):

        q = Queue()

        def callback(in_data, frame_count, time_info, status):
            q.put(in_data)
            return (None, pyaudio.paContinue)

        p.open(
            format=FORMAT,
            channels=CHANNELS,
            rate=SAMPLE_RATE,
            input=True,
            frames_per_buffer=sample_buffer_size,
            stream_callback=callback
        )

        stack = self.instance.exports.stackSave()

        unicodeBytesPointer, getUnicodeBytes = allocate_array_on_stack(
            self.instance,
            [0] * sample_buffer_size
        )

        pre_loop_stack = self.instance.exports.stackSave()
        while True:
            try:
                frame = q.get()

                audioSampleBytesPointer, _ = allocate_array_on_stack(
                    self.instance,
                    frame
                )

                self.instance.exports.quiet_decoder_consume(
                    self.quiet_decoder, audioSampleBytesPointer, sample_buffer_size,
                )

                read = self.instance.exports.quiet_decoder_recv(
                    self.quiet_decoder, unicodeBytesPointer, sample_buffer_size
                )

                if (read != -1):
                    output = bytes(getUnicodeBytes()[:read]).decode(
                        "utf-8", 'ignore')
                    yield output

                self.instance.exports.stackRestore(pre_loop_stack)
            except KeyboardInterrupt:
                break
        self.instance.exports.stackRestore(stack)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['quiet_py_wasm']

package_data = \
{'': ['*']}

install_requires = \
['PyAudio>=0.2.11,<0.3.0',
 'wasmer-compiler-cranelift==1.0.0',
 'wasmer>=1.0.0,<2.0.0']

setup_kwargs = {
    'name': 'quiet-py-wasm',
    'version': '0.1.3',
    'description': 'Python interface to WASM build of Quiet. Work in progress: Current not suitable for use.',
    'long_description': None,
    'author': 'Martin Moxon',
    'author_email': None,
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



import os
from wasmer import engine, Store, Module, Instance, ImportObject, Function, FunctionType, Type
from wasmer_compiler_cranelift import Compiler
import json
import time
from .transmitter import Transmitter
from .receiver import Receiver
from .utils import encode_string

quiet_dir = os.path.dirname(__file__)

quiet_path = os.path.join(quiet_dir, "quiet.wasm")
profiles_path = os.path.join(quiet_dir, "quiet-profiles.json")


def build_instance():
    import_object = ImportObject()

    store = Store(engine.JIT(Compiler))

    import_object.register(
        "env",
        {
            "__sys_getpid": Function(store, lambda: 42, FunctionType([], [Type.I32])),
        }
    )

    import_object.register(
        "wasi_snapshot_preview1",
        {
            "proc_exit": Function(store, lambda *args: None, FunctionType([Type.I32], [])),
            "clock_time_get": Function(store, lambda *args: int(time.time()), FunctionType([Type.I32, Type.I64, Type.I32], [Type.I32])),
            "fd_close": Function(store, lambda *args: 1, FunctionType([Type.I32], [Type.I32])),
            "fd_write": Function(store, lambda *args: 1, FunctionType([Type.I32, Type.I32, Type.I32, Type.I32], [Type.I32])),
            "fd_seek": Function(store, lambda *args: 1, FunctionType([Type.I32, Type.I64, Type.I32, Type.I32], [Type.I32])),
            "fd_read": Function(store, lambda *args: 1, FunctionType([Type.I32, Type.I32, Type.I32, Type.I32], [Type.I32])),
        }
    )

    # Let's compile the module to be able to execute it!
    module = Module(store, open(quiet_path, 'rb').read())

    # Now the module is compiled, we can instantiate it.
    return Instance(module, import_object)


class Quiet:
    def __init__(self):
        self.profiles = json.load(open(profiles_path, 'r'))
        self.instance = build_instance()

    def transmit(self, payload, profile):
        Transmitter(self.instance) \
            .select_profile(self.profiles[profile]) \
            .transmit(encode_string(payload)) \
            .destroy()

    def receive(self, profile):
        return Receiver(self.instance) \
            .select_profile(self.profiles[profile]) \
            .receive()
